
#include <iostream>
#include <fstream>

#include "str.h"
#include "handling.h"
//#include "CFile.h"

typedef unsigned char 	byte;
typedef unsigned short 	word;


#define	C1541_RAM_SIZE				0x800
#define	C1541_RAM_START_BUFFER	1  //darf nicht 0 sein!!!!!! (da buffer 0 fr channel15_buffer) 
#define	C1541_RAM_LAST_BUFFER	7
#define	WORK_BUFFER_SIZE			200

#define	BLOCKS_OF_A_DISK			683
#define	BLOCK_SIZE					0x100
#define	BLOCK_HEADER_SIZE			2
#define	DISK_SIZE					174848
#define  LOGICAL_BAM_POSITION		91392
#define	LOGICAL_DIRECTORY_END_POSITION	96256


#define  SECUNDAR_FLAG_OPEN_FILE			0xf0
#define	SECUNDAR_FLAG_CLOSE_FILE 	 	0xe0
#define	SECUNDAR_FLAG_NORMAL				0x60
#define	DEVICE_FLAG_TALK					0x40
#define	DEVICE_FLAG_LISTEN	  			0x20
#define	DEVICE_FLAG_NONE					0x00


#define	EOF_FLAG							0x8000
#define	FILE_TYPE_BITMASK				0x07
#define FILETYPE_CLOSED_FLAG			0x80
#define FILETYPE_WRITEPROTECTED_FLAG	0x40

#define	FILE_ENTRY_FILENAME_SIZE	16
#define	FILENAME_FILL_BYTE			0xa0

#define	FILE_TYPE_DEL 					0
#define	FILE_TYPE_SEQ 					1
#define	FILE_TYPE_PRG 					2
#define	FILE_TYPE_USR 					3
#define	FILE_TYPE_REL 					4
#define	FILE_TYPE_ANY					7

#define	FILE_ACCESS_READ				100
#define	FILE_ACCESS_WRITE				101
#define	FILE_ACCESS_APPEND			102


#define	DIRECTORY_BASIC_LINE_SIZE		35
#define DIR_EINRUCKEN_FILETYPE			20
#define	BASIC_START						0x1000

#define M_E_INSTRUCTION_PATH							"c:\\z\\C64Connect\\C64Connect\\M-E-Instruction\\"
#define M_E_INSTRUCTION_FLOPPY_FILENAME			"c1541.prg"
#define M_E_INSTRUCTION_C64_FILENAME				"c64.prg"
#define M_E_INSTRUCTION_C64_NEW_FILENAME			"c64new.prg"

#define		ILLEGAL_CHAR					'~'


struct FILE_ENTRY
{
	short				unused;
	byte		filetype;
	byte		track;
	byte		sector;
	char		filename[FILE_ENTRY_FILENAME_SIZE];
	char				rel_file[3];
	long				unused_2;
	short				buffer_klammeraffe;
	word		blocknumber;
};

class BAM
{
	byte*		BAM_pointer;

	public:
		BAM(byte* bam) : BAM_pointer(bam) {}

		BOOL		AllocateBlock(long);
		BOOL		DeAllocBlock(long);
		long		SearchAndAllocFreeBlock(long);
		long		GetFreeBlocks();
};

class Block
{
	public:
		long	track;
		long	sector;
		long	position;

		Block() {}
		Block(long pos) : position(pos) {}
		Block(long pos,byte* diskim) : position(pos), diskimage(diskim) {}
		Block(long track,long sect) { GetLogical(track,sect); }
		Block(long track,long sect,byte* diskim) : diskimage(diskim)
			{ GetLogical(track,sect); }

		void SetDiskImage(byte*);
		void GetPhysical();
		void GetLogical(long,long);
		long GetBlockSize();
		//long GetBlockLength();
		long GetNextBlock();
		void WriteNextBlock(long);
		void WriteEndBlock(long);
		long AddNewBlock();
		long AddNewBlock(long);
		void ClearBlock();

	protected:
		byte* diskimage;

		long NextTrack();
		long NextSector();
		void SetNextTrack(long);
		void SetNextSector(long);
};

class DiskIO : private Block
{
	long		subpos;
	long		block_size;
//	long		read_date_EOF;

	public:
		DiskIO() {}
		DiskIO(long,byte*);

		void SetDiskImage(byte*);
		void SetPosition(long);

		void WriteFileEnd();
		void Write(long);
		long Read();
		long Read_with_EOF();
		void SeekEnd();
};

class Directory : private Block
{
	long localpointer;

	public:
		Directory() {}
		Directory(byte* _diskimage);

		void Reset();
		void SetDiskImage(byte*);

		long GetNextFileEntry();
		void SearchFileEntry();
		BOOL SearchFileEntry(char*);

		long GetFileType();
		void SetFileType(long);
		long GetFilePosition();
		void SetFilePosition(long);
		long GetBlockNumber();
		void SetBlockNumber(long);
		char* GetFileName();
		void SetFileName(char*);
		BOOL CompareFileName(char*);
		void CloseFile();
		void DeleteFile();
		void EnterFilePosition();
		BOOL TestValidFileEntry();
		long CopyFileName(char*);
};

struct SECUNDAR_ADDR_ENTRY
{
	Directory		directory;
	DiskIO			disk_io;
	BOOL				disk_io_read;
	long				buffer_number;
	long				buffer_pointer;
	BOOL				buffer_mode;
	BOOL				directory_mode;
	BOOL				valid;
};


struct c1541_error
{
	char error_text[50];
	long track;
	long sector;
};


class c1541
{
	protected:
		BOOL diskimage_changed;
		BOOL ram_changed;
		byte ram[C1541_RAM_SIZE];

		c1541_error* error;
		long device_flag;
		long bus_status;
		BOOL secundar_instruction_channel;
		long secundar_flag;
		long secundar_addr;
		SECUNDAR_ADDR_ENTRY* secundar;

		byte* work_buffer;
		long work_buffer_pos;
		TextHandling order;

		BOOL overwrite_file;
		//BOOL directory_file;
		//BOOL buffer_mode;
		long file_access;
		long file_type;
		char* filename;

		SECUNDAR_ADDR_ENTRY	secundar_table[16];
		BOOL buffer_alloc_table[C1541_RAM_LAST_BUFFER+1];

		char* directory_filename;
		BOOL directory_start;
		long directory_basic_pointer;
		BOOL directory_end;
		BOOL directory_valid;
		long dir_buffer_pos;
		long dir_line_size;
		char dir_buffer[DIRECTORY_BASIC_LINE_SIZE+10];

		void SetSecundarAddr(long);
		virtual void OpenFile();
		void OpenBuffer();
		virtual void OpenDirectory();
		virtual void CloseFile();
		void DirectoryMode_NextEntry();
		virtual void DirectoryMode_WriteEntry();
		void DirectoryMode_WriteBasicLineNumber(word);
		virtual long ReadDiskIO();
		virtual void WriteDiskIO(long);
		void AnalyseFileName();
		void ExecuteInstruction();
		void Get_C1541_Error();
		unsigned char* Load_C64_File(char*,int&,int&);

		void InstructionIgnoreNumber();
		long InstructionReadChannelNumber();
		long InstructionReadTrackSector();
	
	public:

		byte* diskimage;
		c1541() {}
		virtual void Reset(byte*);
		virtual void Exception(c1541_error*);

		void ListenTalk(long);
	//	void Talk();
		void UnListen();
		void UnTalk();
		void SecundarAddrListen(long);
		void SecundarAddrTalk(long);
		void ReadBus(long);
		long WriteBus();
		byte* BoostWriteBus(long&);
		void BoostReadBus(byte*,long);

		void InstructionBlockTransfer();
		void InstructionBufferPointer();
		void InstructionMemoryWrite();
		void InstructionMemoryExecute();
};

class c1541_extended : public c1541
{
	private:
		int f_access[16];
		int f_type[16];
		byte* f_buffer[16];
		long f_pos[16];
		long f_size[16];
		BOOL directory_last_file;

//		CFileFind filefind;		BRAUCHE NICHT
		str GetFilePath();
	
	protected:
		void OpenFile();
		void OpenDirectory();
		void CloseFile();
		void DirectoryMode_WriteEntry();
		long ReadDiskIO();
		void WriteDiskIO(long);
		int GetFileType(char*);
		char C64_to_PC_ASCII(char);
		char C64_ASCII(char);
		void C64_ASCII(char*);
	
	public:
		str diskimage_path;
		bool extended_flag;

		c1541_extended();
		~c1541_extended();
		void Exception(c1541_error*);
		void Reset(byte*);
	//	byte* BoostWriteBus(long&);
	//	void BoostReadBus(byte*,long);
};

class c1541_diskimage : public c1541_extended
{
	protected:
		void SaveDiskImage();
		void LoadDiskImage();
		//str diskimage_path;

	public:		
		c1541_diskimage();
		~c1541_diskimage() { RemoveDisk(); }
		
		bool UnpackDiskImage(c1541_diskimage*);
		bool AddFileToDiskimage(char*);
		void InsertDisk(char*,bool);
		void RemoveDisk();
		void NewFormatedDisk(str);
		void clrdiskimage();
		void SetDiskimageChanged();
};
